@@ -4,16 +4,17 @@ import sys
 import urllib.parse
 from datetime import date, datetime, timedelta
 from re import match, search
-from typing import Any, Dict, Optional, Set, Tuple
+from typing import Any, Dict, List, Optional, Set, Tuple
 
 from locust.clients import HttpSession, ResponseContextManager
 from requests import Response
 
 from . import logger
+from ._locust_error_handler import log_locust_error, test_response_for_error
 from ._save_request_builder import save_builder
-from .exceptions import BadCredentialsException, MissingCsrfTokenException
-from .helper import (find_component_by_attribute_in_dict, get_username,
-                     log_locust_error, test_response_for_error)
+from .exceptions import BadCredentialsException, MissingCsrfTokenException, ComponentNotFoundException
+from .helper import find_component_by_attribute_in_dict, get_username
+from .records_helper import get_url_stub_from_record_list_post_request_url
 
 log = logger.getLogger(__name__)
 
@@ -62,7 +63,7 @@ class _Interactor:
 
         Examples:
 
-            >>> self.setup_request_headers()
+            >>> self.appian.interactor.setup_request_headers()
         """
 
         uri = uri if uri is not None else self.host
@@ -104,13 +105,19 @@ class _Interactor:
         headers["Accept"] = headers["Accept"] + ", application/json; inlineSail=true; recordHeader=true"
         return headers
 
+    def setup_content_headers(self) -> dict:
+        headers = self.setup_request_headers()
+        headers["Accept"] = "*/*"
+        return headers
+
     def replace_base_path_if_appropriate(self, uri: str) -> str:
         if hasattr(self.client, "base_path_override") and self.client.base_path_override and \
                 self.client.base_path_override != '/suite':
             return uri.replace('/suite', self.client.base_path_override, 1)
         return uri
 
-    def post_page(self, uri: str, payload: Any = {}, headers: Dict[str, Any] = None, label: str = None, files: dict = None, check_login: bool = True) -> Response:
+    def post_page(self, uri: str, payload: Any = {}, headers: Dict[str, Any] = None, label: str = None,
+                  files: dict = None, check_login: bool = True) -> Response:
         """
         Given a uri, executes POST request and returns response
 
@@ -136,11 +143,10 @@ class _Interactor:
         elif isinstance(payload, str):
             post_payload = payload.encode()
         else:
-            log_locust_error(
-                Exception("Cannot POST a payload that is not of type dict or string"),
-                location="_interactor.py/post_page")
+            log_locust_error(Exception("Cannot POST a payload that is not of type dict or string"))
             sys.exit(1)
-        with self.client.post(uri, data=post_payload, headers=headers, name=label, files=files, catch_response=True) as resp:  # type: ResponseContextManager
+        with self.client.post(uri, data=post_payload, headers=headers, name=label, files=files,
+                              catch_response=True) as resp:  # type: ResponseContextManager
             try:
                 test_response_for_error(resp, uri, raise_error=check_login, username=username)
             except Exception as e:
@@ -218,7 +224,8 @@ class _Interactor:
             if login_page_resp.ok and '__appianCsrfToken' in login_page_resp.cookies:
                 self.login()
 
-    def get_page(self, uri: str, headers: Optional[Dict[str, Any]] = None, label: str = None, check_login: bool = True) -> Response:
+    def get_page(self, uri: str, headers: Optional[Dict[str, Any]] = None, label: str = None,
+                 check_login: bool = True) -> Response:
         """
         Given a uri, executes GET request and returns response
 
@@ -247,7 +254,8 @@ class _Interactor:
                 self.write_response_to_lib_folder(label, resp)
             return resp
 
-    def get_webapi(self, uri: str, headers: Dict[str, Any] = None, label: str = None, queryparameters: Dict[str, Any] = {}) -> Response:
+    def get_webapi(self, uri: str, headers: Dict[str, Any] = None, label: str = None,
+                   queryparameters: Dict[str, Any] = {}) -> Response:
         """
         Same as ``get_page``. Additionally it accepts the query parameter to add query parameter while running "GET" operation
         Args:
@@ -258,6 +266,16 @@ class _Interactor:
 
         Returns: Json response of GET operation
 
+        To set custom headers
+
+        >>> headers = self.appian.interactor.setup_request_headers()
+        ... headers['Is-Admin'] = 'true'
+        ... self.appian.interactor.get_webapi('/suite/webapi/headers', headers=headers)
+
+        To set custom query parameters
+
+        >>> params = {'age': 5, 'start-date': '10-05-2020'}
+        ... self.appian.interactor.get_webapi('/suite/webapi/query', queryparameters=params)
         """
         querystring = []
         for k, v in queryparameters.items():
@@ -265,8 +283,6 @@ class _Interactor:
 
         uri += "?" + "&".join(querystring)
         resp = self.get_page(uri, headers=headers, label=label)
-        if self.record_mode:
-            self.write_response_to_lib_folder(label, resp)
         return resp
 
     def upload_document_to_server(self, file_path: str, is_encrypted: bool = False) -> int:
@@ -350,11 +366,7 @@ class _Interactor:
         record_view_url_stub = f"/view/{dashboard}"
         if not record_ref:
             e = Exception("Cannot find _recordRef attribute in RecordLink component.")
-            log_locust_error(
-                e,
-                location="_interactor.py/click_record_link()",
-                raise_error=True
-            )
+            log_locust_error(e, raise_error=True)
         record_link_url_suffix = record_ref + record_view_url_stub
 
         # Logic to construct record link URL in tempo and sites
@@ -362,7 +374,7 @@ class _Interactor:
             record_link_url = "/suite/tempo/records/item/" + record_link_url_suffix
         elif "sites" in get_url and "/record/" in get_url:
             parse_pattern = "/record/"
-            record_link_url = get_url[:get_url.index(parse_pattern)+len(parse_pattern)] + record_link_url_suffix
+            record_link_url = get_url[:get_url.index(parse_pattern) + len(parse_pattern)] + record_link_url_suffix
         elif match(r'.*\/page\/\w+$', get_url):
             record_link_url = get_url + "/record/" + record_link_url_suffix
         # Support record links on site pages
@@ -372,15 +384,12 @@ class _Interactor:
                 page_name = page_search.group()
             else:
                 e = Exception("Unexpected record link URL - couldn't find page name after /pages/")
-                log_locust_error(
-                    e,
-                    location="_interactor.py/click_record_link()",
-                    raise_error=True
-                )
+                log_locust_error(e, raise_error=True)
             parse_pattern = page_name + "/report"
             url_prefix_index = get_url.index(parse_pattern) + len(page_name)
             # record_link_url = get_url[:get_url.index(parse_pattern) + len(page_name)].replace("/pages/",
-            record_link_url = get_url[:url_prefix_index].replace("/pages/", "/page/") + "/record/" + record_link_url_suffix
+            record_link_url = get_url[:url_prefix_index].replace("/pages/",
+                                                                 "/page/") + "/record/" + record_link_url_suffix
         # Support record view links from a record within a site
         elif "record" in get_url:
             site_name = component.get('siteUrlStub', "")
@@ -388,19 +397,11 @@ class _Interactor:
             record_link_url = f"/suite/rest/a/sites/latest/{site_name}/page/{page_name}/record/{record_link_url_suffix}"
         else:
             e = Exception("Unexpected record link URL")
-            log_locust_error(
-                e,
-                location="_interactor.py/click_record_link()",
-                raise_error=True
-            )
+            log_locust_error(e, raise_error=True)
 
         if not get_url or not record_link_url:
             e = Exception("Cannot make Record Link request.")
-            log_locust_error(
-                e,
-                location="_interactor.py/click_record_link()",
-                raise_error=True
-            )
+            log_locust_error(e, raise_error=True)
 
         # Clicking a record link returns a record instance feed - use setup_feed_headers to get the correct headers
         headers = self.setup_feed_headers()
@@ -469,7 +470,8 @@ class _Interactor:
     # COMPONENT RELATED METHODS
 
     def click_component(self, post_url: str, component: Dict[str, Any], context: Dict[str, Any],
-                        uuid: str, label: str = None, headers: Dict[str, Any] = None, client_mode: str = None) -> Dict[str, Any]:
+                        uuid: str, label: str = None, headers: Dict[str, Any] = None,
+                        client_mode: str = None) -> Dict[str, Any]:
         '''
             Calls the post operation to click certain SAIL components such as Buttons and Dynamic Links
 
@@ -488,10 +490,10 @@ class _Interactor:
             component = component["link"]
             component["label"] = wrapper_label
 
-        payload = save_builder()\
-            .component(component)\
-            .context(context)\
-            .uuid(uuid)\
+        payload = save_builder() \
+            .component(component) \
+            .context(context) \
+            .uuid(uuid) \
             .build()
 
         locust_label = label or f'Click \'{component["label"]}\' Component'
@@ -505,8 +507,8 @@ class _Interactor:
     click_button = click_component
     click_link = click_component
 
-    def send_dropdown_update(self, post_url: str, dropdown: Dict[str, Any],
-                             context: Dict[str, Any], uuid: str, index: int, label: str = None) -> Dict[str, Any]:
+    def send_dropdown_update(self, post_url: str, dropdown: Dict[str, Any], context: Dict[str, Any],
+                             uuid: str, index: int, label: str = None, url_stub: str = None) -> Dict[str, Any]:
         '''
             Calls the post operation to send an update to a dropdown
 
@@ -518,6 +520,7 @@ class _Interactor:
                 index: location of the dropdown value
                 label: the label to be displayed by locust for this action
                 headers: header for the REST API call
+                url_stub: the URL stub for the page; only required to interact with user filter dropdowns on a record list
 
             Returns: the response of post operation as json
         '''
@@ -525,12 +528,14 @@ class _Interactor:
             "#t": "Integer",
             "#v": index
         }
-        payload = save_builder()\
-            .component(dropdown)\
-            .context(context)\
-            .uuid(uuid)\
-            .value(new_value)\
-            .record_url_stub(self._get_record_instance_list_url_stub(post_url))\
+        # url_stub should only be populated if the page is a record list
+        url_stub = url_stub or get_url_stub_from_record_list_post_request_url(post_url)
+        payload = save_builder() \
+            .component(dropdown) \
+            .context(context) \
+            .uuid(uuid) \
+            .value(new_value) \
+            .record_url_stub(url_stub) \
             .build()
 
         locust_label = label or f'Select \'{dropdown["label"]}\' Dropdown'
@@ -540,6 +545,44 @@ class _Interactor:
         )
         return resp.json()
 
+    def send_multiple_dropdown_update(self, post_url: str, multi_dropdown: Dict[str, Any], context: Dict[str, Any],
+                                      uuid: str, index: List[int], label: str = None, url_stub: str = None) -> Dict[str, Any]:
+        '''
+            Calls the post operation to send an update to a multiple dropdown
+
+            Args:
+                post_url: the url (not including the host and domain) to post to
+                dropdown: the JSON code for the desired dropdown
+                context: the Sail context parsed from the json response
+                uuid: the uuid parsed from the json response
+                index: locations of the multiple dropdown value
+                label: the label to be displayed by locust for this action
+                headers: header for the REST API call
+                url_stub: the URL stub for the page; only required to interact with user filter dropdowns on a record list
+
+            Returns: the response of post operation as json
+        '''
+        new_value = {
+            "#t": "Integer?list",
+            "#v": index
+        }
+        # url_stub should only be populated if the page is a record list
+        url_stub = url_stub or get_url_stub_from_record_list_post_request_url(post_url)
+        payload = save_builder() \
+            .component(multi_dropdown) \
+            .context(context) \
+            .uuid(uuid) \
+            .value(new_value) \
+            .record_url_stub(url_stub) \
+            .build()
+
+        locust_label = label or f'Select \'{multi_dropdown["label"]}\' Dropdown'
+
+        resp = self.post_page(
+            self.host + post_url, payload=payload, label=locust_label
+        )
+        return resp.json()
+
     def get_primary_button_payload(self, page_content_in_json: Dict[str, Any]) -> Dict[str, Any]:
         """
         Finds the primary button from the page content response and creates the payload which can be used to simulate a click
@@ -554,10 +597,10 @@ class _Interactor:
         primary_button["#t"] = "ButtonWidget"
         context = page_content_in_json["context"]
         uuid = page_content_in_json["uuid"]
-        payload = save_builder()\
-            .component(primary_button)\
-            .context(context)\
-            .uuid(uuid)\
+        payload = save_builder() \
+            .component(primary_button) \
+            .context(context) \
+            .uuid(uuid) \
             .build()
 
         return payload
@@ -578,11 +621,11 @@ class _Interactor:
 
         """
         new_value = {"#t": "Text", "#v": text}
-        payload = save_builder()\
-            .component(text_field)\
-            .context(context)\
-            .uuid(uuid)\
-            .value(new_value)\
+        payload = save_builder() \
+            .component(text_field) \
+            .context(context) \
+            .uuid(uuid) \
+            .value(new_value) \
             .build()
 
         locust_label = label or f'Fill \'{text_field["label"]}\' TextField'
@@ -612,11 +655,11 @@ class _Interactor:
             "typedText": text
         }
 
-        payload = save_builder()\
-            .component(picker_field)\
-            .context(context)\
-            .uuid(uuid)\
-            .value(new_value)\
+        payload = save_builder() \
+            .component(picker_field) \
+            .context(context) \
+            .uuid(uuid) \
+            .value(new_value) \
             .build()
 
         locust_label = label or f'Fill \'{picker_field["label"]}\' PickerField'
@@ -649,11 +692,11 @@ class _Interactor:
             "identifiers": identifiers_list
         }
 
-        payload = save_builder()\
-            .component(picker_field)\
-            .context(context)\
-            .uuid(uuid)\
-            .value(new_value)\
+        payload = save_builder() \
+            .component(picker_field) \
+            .context(context) \
+            .uuid(uuid) \
+            .value(new_value) \
             .build()
 
         locust_label = label or f'Fill \'{picker_field["label"]}\' PickerField'
@@ -664,7 +707,8 @@ class _Interactor:
         return resp.json()
 
     def select_checkbox_item(self, post_url: str, checkbox: Dict[str, Any],
-                             context: Dict[str, Any], uuid: str, indices: list, context_label: str = None) -> Dict[str, Any]:
+                             context: Dict[str, Any], uuid: str, indices: list,
+                             context_label: str = None) -> Dict[str, Any]:
         '''
             Calls the post operation to send an update to a checkbox to check all appropriate boxes
 
@@ -683,14 +727,15 @@ class _Interactor:
             "#t": "Integer?list",
             "#v": indices if indices else None
         }
-        payload = save_builder()\
-            .component(checkbox)\
-            .context(context)\
-            .uuid(uuid)\
-            .value(new_value)\
+        payload = save_builder() \
+            .component(checkbox) \
+            .context(context) \
+            .uuid(uuid) \
+            .value(new_value) \
             .build()
 
-        locust_label = context_label or "Checking boxes for " + checkbox.get("testLabel", checkbox.get("label", "label-not-found"))
+        locust_label = context_label or "Checking boxes for " + checkbox.get("testLabel",
+                                                                             checkbox.get("label", "label-not-found"))
 
         resp = self.post_page(
             self.host + post_url, payload=payload, label=locust_label
@@ -727,11 +772,11 @@ class _Interactor:
         else:
             raise Exception(f"Cannot click a tab with label: '{tab_label}' inside the TabButtonGroup component")
 
-        payload = save_builder()\
-            .component(tab_group_component)\
-            .context(context)\
-            .uuid(uuid)\
-            .value(new_value)\
+        payload = save_builder() \
+            .component(tab_group_component) \
+            .context(context) \
+            .uuid(uuid) \
+            .value(new_value) \
             .build()
 
         locust_label = f"Selecting tab with label: '{tab_label}' inside TabButtonGroup component"
@@ -761,11 +806,11 @@ class _Interactor:
             "#t": "Integer",
             "#v": index
         }
-        payload = save_builder()\
-            .component(buttons)\
-            .context(context)\
-            .uuid(uuid)\
-            .value(new_value)\
+        payload = save_builder() \
+            .component(buttons) \
+            .context(context) \
+            .uuid(uuid) \
+            .value(new_value) \
             .build()
 
         resp = self.post_page(
@@ -795,11 +840,11 @@ class _Interactor:
             "#t": "CollaborationDocument",
             "id": doc_id
         }
-        payload = save_builder()\
-            .component(upload_field)\
-            .context(context)\
-            .uuid(uuid)\
-            .value(new_value)\
+        payload = save_builder() \
+            .component(upload_field) \
+            .context(context) \
+            .uuid(uuid) \
+            .value(new_value) \
             .build()
 
         locust_label = locust_label or "Uploading Document to " + \
@@ -832,11 +877,11 @@ class _Interactor:
             "#v": f"{date_input.isoformat()}Z" if date_input else None
         }
 
-        payload = save_builder()\
-            .component(date_field_component)\
-            .context(context)\
-            .uuid(uuid)\
-            .value(new_value)\
+        payload = save_builder() \
+            .component(date_field_component) \
+            .context(context) \
+            .uuid(uuid) \
+            .value(new_value) \
             .build()
 
         locust_label = locust_label or "Filling Date Field for " + \
@@ -870,11 +915,11 @@ class _Interactor:
             "#v": f"{datetime_input.replace(second=0, microsecond=0).isoformat()}Z" if datetime_input else None
         }
 
-        payload = save_builder()\
-            .component(datetime_field)\
-            .context(context)\
-            .uuid(uuid)\
-            .value(new_value)\
+        payload = save_builder() \
+            .component(datetime_field) \
+            .context(context) \
+            .uuid(uuid) \
+            .value(new_value) \
             .build()
 
         locust_label = locust_label or "Filling Date Time Field for " + \
@@ -904,11 +949,11 @@ class _Interactor:
 
             Returns: the response of post operation as jso
         """
-        payload = save_builder()\
-            .component(grid_component)\
-            .context(context)\
-            .uuid(uuid)\
-            .value(new_grid_save_value)\
+        payload = save_builder() \
+            .component(grid_component) \
+            .context(context) \
+            .uuid(uuid) \
+            .value(new_grid_save_value) \
             .build()
 
         locust_label = context_label or "Updating Grid " + grid_component.get("label", "")
@@ -934,11 +979,11 @@ class _Interactor:
             Returns: the response of post operation as json
         """
         url_stub = post_url.split('/')[-1]
-        payload = save_builder()\
-            .component(grid_component)\
-            .context(context)\
-            .uuid(uuid)\
-            .record_url_stub(url_stub)\
+        payload = save_builder() \
+            .component(grid_component) \
+            .context(context) \
+            .uuid(uuid) \
+            .record_url_stub(url_stub) \
             .build()
 
         locust_label = context_label or "Updating Record Grid " + grid_component.get("label", "")
@@ -948,18 +993,93 @@ class _Interactor:
         resp.raise_for_status()
         return resp.json()
 
-    def _get_record_instance_list_url_stub(self, post_url: str) -> Optional[str]:
+    def refresh_after_record_action(self, post_url: str, record_action_component: Dict[str, Any],
+                                    record_action_trigger_component: Dict[str, Any],
+                                    context: Dict[str, Any], uuid: str, label: str = None) -> Dict[str, Any]:
         """
-            Given post_url, returns the URL stub IF the url matches the url for a record instance list.
-            If, not returns None.
+            Calls the post operation to refresh a form after completion of a record action
 
             Args:
                 post_url: the url (not including the host and domain) to post to
+                record_action_component: the JSON representing the relevant record action component
+                record_action_trigger_component: the JSON representing the form's record action trigger component
+                context: the Sail context parsed from the json response
+                uuid: the uuid parsed from the json response
 
-            Returns: The url stub if post_url matches a record instance list url, otherwise None
+            Returns: the response of post operation as json
         """
-        record_url_match = match(r'[\S]+\/pages\/records\/recordType\/([\w]+)', post_url)
-        return record_url_match.groups()[0] if record_url_match else None
+        # Get the payload for the record action on submit
+        record_action_payload = save_builder() \
+            .component(record_action_component) \
+            .context(context) \
+            .uuid(uuid) \
+            .value(dict()) \
+            .build()
+
+        # Get the payload for the record action trigger
+        record_action_trigger_payload = save_builder() \
+            .component(record_action_trigger_component) \
+            .context(context) \
+            .uuid(uuid) \
+            .value(dict()) \
+            .build()
+
+        # Get both save requests
+        record_action_save_request = record_action_payload["updates"]["#v"][0]
+        record_action_trigger_save_request = record_action_trigger_payload["updates"]["#v"][0]
+
+        # Update the main payload with the both save requests
+        record_action_payload["updates"]["#v"] = [record_action_save_request, record_action_trigger_save_request]
+
+        resp = self.post_page(
+            self.host + post_url, payload=record_action_payload, label=label
+        )
+        return resp.json()
+
+    def click_record_search_button(self, post_url: str, component: Dict[str, Any], context: Dict[str, Any],
+                                   uuid: str, label: str = None) -> Dict[str, Any]:
+        """
+            Calls the post operation to click a record search button
+
+            Args:
+                post_url: the url (not including the host and domain) to post to
+                component: the JSON code for the desired SearchBoxWidget component
+                context: the Sail context parsed from the json response
+                uuid: the uuid parsed from the json response
+                label: the label to be displayed by locust for this action
+
+            Returns: the response of post operation as json
+        """
+
+        # Create a new ButtonWidget component from the SearchBoxWidget
+        c_id = component["_cId"]
+        action = find_component_by_attribute_in_dict("testLabel", "Applications-searchLink", component)
+        if not action:
+            raise ComponentNotFoundException(
+                f'''Could not find the Applications-searchLink component in the provided component
+                ''')
+        save_into = action["saveInto"]
+
+        search_box_button_component = {
+            "_cId": f"{c_id}_buttonWidget",
+            "value": None,
+            "saveInto": save_into,
+            "saveType": "PRIMARY",
+            "#t": "ButtonWidget"
+        }
+
+        payload = save_builder() \
+            .component(search_box_button_component) \
+            .context(context) \
+            .uuid(uuid) \
+            .build()
+
+        locust_label = label or f'Click \'{component["searchButtonLabel"]}\' Component'
+
+        resp = self.post_page(
+            self.host + post_url, payload=payload, label=locust_label
+        )
+        return resp.json()
 
 
 class DataTypeCache(object):
