@@ -5,8 +5,7 @@ import json
 import os
 import random
 import warnings
-from functools import wraps
-from typing import Any, Callable, Dict, List, Optional
+from typing import Any, Dict, List, Optional
 from urllib.parse import quote, urlparse
 
 from appian_locust.records_helper import _is_grid
@@ -14,15 +13,21 @@ from appian_locust.records_helper import _is_grid
 from . import logger
 from ._grid_interactor import GridInteractor
 from ._interactor import _Interactor
+from ._locust_error_handler import raises_locust_error
+from ._task_opener import _TaskOpener
 from ._ui_reconciler import UiReconciler
+from .exceptions import ComponentNotFoundException, InvalidComponentException, ChoiceNotFoundException
 from .helper import (extract_all_by_label, find_component_by_attribute_in_dict,
                      find_component_by_index_in_dict,
-                     find_component_by_label_and_type_dict, log_locust_error)
+                     find_component_by_label_and_type_dict)
 from .records_helper import (get_record_header_response,
-                             get_record_summary_view_response)
+                             get_record_summary_view_response,
+                             get_url_stub_from_record_list_url_path)
 
 KEY_UUID = "uuid"
 KEY_CONTEXT = "context"
+START_PROCESS_LINK_TYPE = 'StartProcessLink'
+PROCESS_TASK_LINK_TYPE = 'ProcessTaskLink'
 COMPONENTS_THAT_CAN_BE_FILLED = ["ParagraphField", "TextField"]
 
 log = logger.getLogger(__name__)
@@ -33,19 +38,6 @@ class ClientMode(enum.Enum):
     DESIGN = 'DESIGN'
 
 
-def raises_locust_error(location: str) -> Callable:
-    def should_log_loc_error(func: Callable) -> Callable:
-        @wraps(func)
-        def func_wrapper(*args: Any, **kwargs: Any) -> Optional[Callable]:
-            try:
-                return func(*args, **kwargs)
-            except Exception as e:
-                log_locust_error(e, location=location, raise_error=True)
-                return None
-        return func_wrapper
-    return should_log_loc_error
-
-
 class SailUiForm:
     def __init__(self, interactor: _Interactor, state: Dict[str, Any], url: str, breadcrumb: str = "SailUi"):
         """
@@ -60,6 +52,7 @@ class SailUiForm:
 
         """
         self.interactor: _Interactor = interactor
+        self.task_opener: _TaskOpener = _TaskOpener(self.interactor)
         self.state: Dict[str, Any] = state
         self.form_url = url
         if any(key not in self.state for key in (KEY_CONTEXT, KEY_UUID)):
@@ -110,8 +103,8 @@ class SailUiForm:
 
     # TODO: Handle components on a page with the same label
 
-    @raises_locust_error("uiform.py/fill_text_field()")
-    def fill_text_field(self, label: str, value: str, locust_request_label: str = "") -> 'SailUiForm':
+    @raises_locust_error
+    def fill_text_field(self, label: str, value: str, is_test_label: bool = False, locust_request_label: str = "") -> 'SailUiForm':
         """
         Fills a field on the form, if there is one present with the following label (case sensitive)
         Otherwise throws a NotFoundException
@@ -119,6 +112,7 @@ class SailUiForm:
         Args:
             label(str): Label of the field to fill out
             value(str): Value to update the label to
+            is_test_label(bool): If you are filling a text field via a test label instead of a label, set this boolean to true
 
         Keyword Args:
             locust_request_label(str): Label used to identify the request for locust statistics
@@ -130,8 +124,9 @@ class SailUiForm:
             >>> form.fill_text_field('Title','My New Novel')
 
         """
+        attribute_to_find = 'testLabel' if is_test_label else 'label'
         component = find_component_by_attribute_in_dict(
-            'label', label, self.state)
+            attribute_to_find, label, self.state)
         self._validate_component_found(component, label)
 
         reeval_url = self._get_update_url_for_reeval(self.state)
@@ -143,7 +138,7 @@ class SailUiForm:
 
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/fill_field_by_index()")
+    @raises_locust_error
     def fill_field_by_index(self, type_of_component: str, index: int, text_to_fill: str, locust_request_label: str = "") -> 'SailUiForm':
         """
         Selects a Field by its index and fills it with a text value
@@ -178,7 +173,7 @@ class SailUiForm:
 
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/fill_field_by_any_attribute()")
+    @raises_locust_error
     def fill_field_by_any_attribute(self, attribute: str, value_for_attribute: str, text_to_fill: str, locust_request_label: str = "") -> 'SailUiForm':
         """
         Selects a Field by "attribute" and its value provided "value_for_attribute"
@@ -228,8 +223,9 @@ class SailUiForm:
     # Aliases for fill_text_field() function
     fill_paragraph_field = fill_text_field
 
-    @raises_locust_error("uiform.py/fill_picker_field()")
-    def fill_picker_field(self, label: str, value: str, fill_request_label: str = "", pick_request_label: str = "") -> 'SailUiForm':
+    @raises_locust_error
+    def fill_picker_field(self, label: str, value: str, identifier: str = 'id',
+                          fill_request_label: str = "", pick_request_label: str = "") -> 'SailUiForm':
         """
         Enters the value in the picker widget and selects one of the suggested items
         if the widget is present with the following label (case sensitive)
@@ -245,6 +241,7 @@ class SailUiForm:
             value(str): Value to update the label to
 
         Keyword Args:
+            identifier(str): Key to select the field to filter on, defaults to 'id'
             fill_request_label(str): Label to associate in locust statistics with filling the picker field
             pick_request_label(str): Label to associate in locust statistics with selecting the picker suggestion
 
@@ -254,6 +251,7 @@ class SailUiForm:
 
             >>> form.fill_picker_field('Title','My New Novel')
             >>> form.fill_picker_field('People','Jeff George')
+            >>> form.fill_picker_field('Customer', 'GAC Guyana', identifier='code')
 
         """
         # pickerFieldCustom will add a test-Label at the level where the suggestions/saveInto exist
@@ -280,14 +278,17 @@ class SailUiForm:
             raise Exception(f"No identifiers found when '{value}' was entered in the picker field.")
 
         # Introspect to see if there's an ID
-        index_by_id = identifiers[0].get('id') is not None
-        id_index = 'id' if index_by_id else '#v'
+        index_by_id = identifiers[0].get(identifier) is not None
+        id_index = identifier if index_by_id else '#v'
+
         v_or_id = [identifier.get(id_index) for identifier in identifiers if identifier.get(id_index)]
-        t = [identifier.get('#t') for identifier in identifiers]
+
         if not v_or_id:
             raise Exception(f"Could not extract picker values '{id_index}' from suggestions_list {suggestions_list}")
+
         v_choice = random.choice(range(len(v_or_id)))
-        dict_value = {id_index: v_or_id[v_choice], "#t": t[v_choice]}
+
+        dict_value = identifiers[v_choice]
 
         locust_label = pick_request_label or f"{self.breadcrumb}.SelectPickerSuggestion.{label}"
         newer_state = self.interactor.select_pickerfield_suggestion(
@@ -296,14 +297,18 @@ class SailUiForm:
         if not newer_state:
             raise Exception(f"No response returned when trying to update field with label '{label}'")
 
-        return self._reconcile_state(new_state)
+        return self._reconcile_state(newer_state)
 
-    @raises_locust_error("uiform.py/click()")
+    @raises_locust_error
     def click(self, label: str, is_test_label: bool = False, locust_request_label: str = "") -> 'SailUiForm':
         """
         Clicks on a component on the form, if there is one present with the following label (case sensitive)
         Otherwise throws a NotFoundException
 
+        Can also be called as 'click_link' or 'click_button' to convey intent
+
+        This can also click StartProcessLinks or ProcessTaskLinks
+
         Args:
             label(str): Label of the component to click
             is_test_label(bool): If you are clicking a button or link via a test label instead of a label, set this boolean to true
@@ -318,26 +323,80 @@ class SailUiForm:
             >>> form.click('Submit')
             >>> form.click('SampleTestLabel', is_test_label = True)
 
+        """
+        locust_label = locust_request_label or f"{self.breadcrumb}.Click.{label}"
+        return self._click(label, is_test_label=is_test_label, locust_request_label=locust_label)
+
+    @raises_locust_error
+    def click_button(self, label: str, is_test_label: bool = False, locust_request_label: str = "") -> 'SailUiForm':
+        """
+        Clicks on a component on the form, if there is one present with the following label (case sensitive)
+        Otherwise throws a NotFoundException
+
+        This can also click StartProcessLinks or ProcessTaskLinks
+
+        Args:
+            label(str): Label of the component to click
+            is_test_label(bool): If you are clicking a button via a test label instead of a label, set this boolean to true
+
+        Keyword Args:
+            locust_request_label(str): Label used to identify the request for locust statistics
+
+        Returns (SailUiForm): The latest state of the UiForm
+
+        Examples:
+
+            >>> form.click_button('Save')
+            >>> form.click_link('Update')
+
+        """
+        locust_label = locust_request_label or f"{self.breadcrumb}.ClickButton.{label}"
+        return self._click(label, is_test_label=is_test_label, locust_request_label=locust_label)
+
+    @raises_locust_error
+    def click_link(self, label: str, is_test_label: bool = False, locust_request_label: str = "") -> 'SailUiForm':
+        """
+        Clicks on a component on the form, if there is one present with the following label (case sensitive)
+        Otherwise throws a NotFoundException
+
+        This can also click StartProcessLinks or ProcessTaskLinks
+
+        Args:
+            label(str): Label of the component to click
+            is_test_label(bool): If you are clicking a link via a test label instead of a label, set this boolean to true
+
+        Keyword Args:
+            locust_request_label(str): Label used to identify the request for locust statistics
+
+        Returns (SailUiForm): The latest state of the UiForm
+
+        Examples:
+
+            >>> form.click_link('Update')
+
+        """
+        locust_label = locust_request_label or f"{self.breadcrumb}.ClickLink.{label}"
+        return self._click(label, is_test_label=is_test_label, locust_request_label=locust_label)
+
+    def _click(self, label: str, is_test_label: bool = False, locust_request_label: str = "") -> 'SailUiForm':
+        """
+        Internal function wrapped by various click methods
         """
         attribute_to_find = 'testLabel' if is_test_label else 'label'
-        component = find_component_by_attribute_in_dict(
-            attribute_to_find, label, self.state)
+
+        component = find_component_by_attribute_in_dict(attribute_to_find, label, self.state)
+
         self._validate_component_found(component, label)
 
         locust_label = locust_request_label or f"{self.breadcrumb}.Click.{label}"
         reeval_url = self._get_update_url_for_reeval(self.state)
-        new_state = self.interactor.click_component(
-            reeval_url, component, self.context, self.uuid, label=locust_label)
+        new_state = self._dispatch_click(component=component, locust_label=locust_label)
 
         if not new_state:
             raise Exception(f"No response returned when trying to click button with label '{label}'")
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    # Aliases for click() function
-    click_button = click
-    click_link = click
-
-    @raises_locust_error("uiform.py/click_card_layout_by_index()")
+    @raises_locust_error
     def click_card_layout_by_index(self, index: int, locust_request_label: str = "") -> 'SailUiForm':
         """
         Clicks a card layout link by index.
@@ -362,13 +421,11 @@ class SailUiForm:
         """
         component = find_component_by_index_in_dict("CardLayout", index, self.state)
 
-        if "link" in component and component["link"]:
-            link_component = component["link"]
-        else:
+        if not component.get("link"):
             raise Exception(f"CardLayout found at index: {index} does not have a link on it")
 
+        link_component = component.get('link')
         locust_label = locust_request_label or f"{self.breadcrumb}.ClickCardLayout.Index.{index}"
-
         if link_component["#t"] == "StartProcessLink":
             site_name = link_component["siteUrlStub"] or "D6JMim"
             page_name = link_component["sitePageUrlStub"]
@@ -382,14 +439,15 @@ class SailUiForm:
         reeval_url = self._get_update_url_for_reeval(new_state)
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/click_record_link()")
-    def click_record_link(self, label: str, locust_request_label: str = "") -> 'SailUiForm':
+    @raises_locust_error
+    def click_record_link(self, label: str, is_test_label: bool = False, locust_request_label: str = "") -> 'SailUiForm':
         """
         Click a record link on the form if there is one present with the following label (case sensitive)
         Otherwise throws a ComponentNotFoundException
 
         Args:
             label(str): Label of the record link to click
+            is_test_label(bool): If you are clicking a record link via a test label instead of a label, set this boolean to true
 
         Keyword Args:
             locust_request_label(str): Label used to identify the request for locust statistics
@@ -397,7 +455,8 @@ class SailUiForm:
         Returns (SailUiForm): The record form (feed) for the linked record.
 
         """
-        component = find_component_by_attribute_in_dict('label', label, self.state)
+        attribute_to_find = 'testLabel' if is_test_label else 'label'
+        component = find_component_by_attribute_in_dict(attribute_to_find, label, self.state)
         self._validate_component_found(component, label)
 
         reeval_url = self._get_update_url_for_reeval(self.state)
@@ -406,7 +465,7 @@ class SailUiForm:
                                                       locust_label=locust_label)
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/click_start_process_link()")
+    @raises_locust_error
     def click_start_process_link(self, label: str, site_name: str, page_name: str, is_mobile: bool = False, locust_request_label: str = "") -> 'SailUiForm':
         """
         Clicks a start process link on the form by label
@@ -428,7 +487,7 @@ class SailUiForm:
 
         """
 
-        component = find_component_by_label_and_type_dict('label', label, 'StartProcessLink', self.state)
+        component = find_component_by_label_and_type_dict('label', label, START_PROCESS_LINK_TYPE, self.state)
         self._validate_component_found(component, label)
 
         locust_label = locust_request_label or f"{self.breadcrumb}.ClickStartProcessLink.{label}"
@@ -438,7 +497,7 @@ class SailUiForm:
         reeval_url = self._get_update_url_for_reeval(new_state)
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/click_start_process_link_on_mobile()")
+    @raises_locust_error
     def click_start_process_link_on_mobile(self, label: str, site_name: str, page_name: str, locust_request_label: str = "") -> 'SailUiForm':
         """
         Clicks a start process link on the form by label (for Mobile)
@@ -479,7 +538,50 @@ class SailUiForm:
                                                         cache_key=cache_key, site_name=site_name, page_name=page_name, is_mobile=is_mobile,
                                                         locust_request_label=locust_request_label)
 
-    @raises_locust_error("uiform.py/click_related_action_link()")
+    def _dispatch_click(self, component: Dict[str, Any], locust_label: str) -> Dict[str, Any]:
+        """
+        Dispatches the appropriate link interaction based on the link type if appropriate
+
+        Args:
+            link_component (Dict[str, Any]): Link component to interact with
+            locust_label (str): Label used to identify the request for locust statistics
+
+        Returns:
+            Dict[str, Any]: State returned by the interaction
+        """
+        component_type = component.get('#t')
+        # Check if component is already a supported link
+        if component_type in [START_PROCESS_LINK_TYPE, PROCESS_TASK_LINK_TYPE]:
+            link_component = component
+            link_type = component_type
+        else:
+            link_component = component.get('link', {})
+            link_type = link_component.get('#t')
+
+        if link_type == START_PROCESS_LINK_TYPE:
+            site_name = link_component["siteUrlStub"] or "D6JMim"
+            page_name = link_component["sitePageUrlStub"]
+            new_state = self._click_start_process_link(site_name, page_name, False, link_component, locust_request_label=locust_label)
+        elif link_type == PROCESS_TASK_LINK_TYPE:
+            task_name = link_component.get('label') or 'Unnammed Task'
+            task_id = link_component.get('opaqueTaskId')
+            if not task_id:
+                raise Exception(f"No task id found for task with name '{task_name}'")
+            site_name = link_component.get("siteUrlStub") or "D6JMim"
+            page_name = link_component.get("sitePageUrlStub")
+            headers = {
+                'X-Site-UrlStub': site_name,
+                'X-Page-UrlStub': page_name,
+                'X-Client-Mode': 'SITES'
+            }
+            new_state = self.task_opener.visit_by_task_id(task_name, task_id, extra_headers=headers)
+        elif link_component:
+            new_state = self.interactor.click_component(self.form_url, link_component, self.context, self.uuid, label=locust_label)
+        else:
+            new_state = self.interactor.click_component(self.form_url, component, self.context, self.uuid, label=locust_label)
+        return new_state
+
+    @raises_locust_error
     def click_related_action(self, label: str, locust_request_label: str = "") -> 'SailUiForm':
         """
         Clicks a related action (either a related action button or link) on the form by label
@@ -527,7 +629,36 @@ class SailUiForm:
         reeval_url = self._get_update_url_for_reeval(new_state)
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/select_dropdown_item()")
+    @raises_locust_error
+    def get_dropdown_items(self, label: str, is_test_label: bool = False) -> List[str]:
+        """
+        Gets all dropdown items for the dropdown label provided on the form
+        If no dropdown found, throws a NotFoundException
+
+        Args:
+            label(str): Label of the dropdown
+            is_test_label(bool): If you are interacting with a dropdown via a test label instead of a label, set this boolean to true.
+                                 User filters on a record instance list use test labels.
+
+        Returns (List): A list of all the choices in the dropdown
+
+        Examples:
+
+            >>> form.get_dropdown_items('MyDropdown')
+
+        """
+        attribute_to_find = 'testLabel' if is_test_label else 'label'
+        component = find_component_by_attribute_in_dict(
+            attribute_to_find, label, self.state)
+
+        self._validate_component_found(component, label)
+
+        choices: list = component.get('choices')
+        if choices is None or not isinstance(choices, list):
+            raise InvalidComponentException(f"No choices found for component {label}, is the component a Dropdown?")
+        return choices
+
+    @raises_locust_error
     def select_dropdown_item(self, label: str, choice_label: str, locust_request_label: str = "", is_test_label: bool = False) -> 'SailUiForm':
         """
         Selects a dropdown item on the form
@@ -566,8 +697,65 @@ class SailUiForm:
 
         locust_label = locust_request_label or f'{self.breadcrumb}.SelectDropdownWithLabel.{choice_label}'
         reeval_url = self._get_update_url_for_reeval(self.state)
+
+        # Opting to use this field, rather than self.form_url, because 'sail-application-url' is the same between web and mobile
+        url = self.state.get('sail-application-url')
+        # url_stub should only be populated if the page is a record list
+        url_stub = get_url_stub_from_record_list_url_path(url)
+
         new_state = self.interactor.send_dropdown_update(
-            reeval_url, component, self.context, self.uuid, index=index, label=locust_label)
+            reeval_url, component, self.context, self.uuid, index=index, label=locust_label, url_stub=url_stub)
+        if not new_state:
+            raise Exception(
+                f"No response returned when trying to click button with label '{label}'")
+
+        return self._reconcile_state(new_state, form_url=reeval_url)
+
+    @raises_locust_error
+    def select_multi_dropdown_item(self, label: str, choice_label: List[str], locust_request_label: str = "", is_test_label: bool = False) -> 'SailUiForm':
+        """
+        Selects a multiple dropdown item on the form
+        If no multiple dropdown found, throws a NotFoundException
+        If no element found, throws a ChoiceNotFoundException
+
+        Args:
+            label(str): Label of the dropdown
+            choice_label([str]): Label(s) of the multiple dropdown item to select
+            is_test_label(bool): If you are interacting with a multiple dropdown via a test label instead of a label, set this boolean to true.
+                                 User filters on a record instance list use test labels.
+
+        Keyword Args:
+            locust_request_label(str): Label used to identify the request for locust statistics
+
+        Returns (SailUiForm): The latest state of the UiForm
+
+        Examples:
+
+            >>> form.select_multi_dropdown_item('MyMultiDropdown', ['My First Choice','My Second Choice'])
+
+        """
+        attribute_to_find = 'testLabel' if is_test_label else 'label'
+        component = find_component_by_attribute_in_dict(
+            attribute_to_find, label, self.state)
+
+        self._validate_component_found(component, label)
+
+        choices: list = component.get('choices')
+        if not choices:
+            raise InvalidComponentException(f"No choices found for component {label}, is the component a Dropdown?")
+        if not isinstance(choice_label, list) and choice_label not in choices:
+            raise ChoiceNotFoundException(f"Choice {choice_label} not found for component {label}, valid choices were {choices}")
+        index_multi = [choices.index(current_label) + 1 for current_label in choice_label]  # Appian is _sigh_ one indexed
+        locust_label = locust_request_label or f'{self.breadcrumb}.SelectMultupleDropdownWithLabel.{choice_label}'
+        reeval_url = self._get_update_url_for_reeval(self.state)
+
+        # Opting to use this field, rather than self.form_url, because 'sail-application-url' is the same between web and mobile
+        url = self.state.get('sail-application-url')
+        # url_stub should only be populated if the page is a record list
+        url_stub = get_url_stub_from_record_list_url_path(url)
+
+        new_state = self.interactor.send_multiple_dropdown_update(
+            reeval_url, component, self.context, self.uuid, index=index_multi, label=locust_label, url_stub=url_stub)
         if not new_state:
             raise Exception(
                 f"No response returned when trying to click button with label '{label}'")
@@ -604,7 +792,7 @@ class SailUiForm:
 
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/check_checkbox_by_test_label()")
+    @raises_locust_error
     def check_checkbox_by_test_label(self, test_label: str, indices: List[int], locust_request_label: str = "") -> 'SailUiForm':
         """
         Checks a checkbox by its testLabel attribute
@@ -631,7 +819,7 @@ class SailUiForm:
         locust_label = locust_request_label or f'{self.breadcrumb}.CheckCheckboxByTestLabel.{test_label}'
         return self._check_checkbox_by_attribute('testLabel', test_label, indices, locust_request_label=locust_label)
 
-    @raises_locust_error("uiform.py/check_checkbox_by_label()")
+    @raises_locust_error
     def check_checkbox_by_label(self, label: str, indices: List[int], locust_request_label: str = "") -> 'SailUiForm':
         """
         Checks a checkbox by its label
@@ -658,7 +846,7 @@ class SailUiForm:
         locust_label = locust_request_label or f'{self.breadcrumb}.CheckCheckboxByLabel.{label}'
         return self._check_checkbox_by_attribute('label', label, indices, locust_request_label=locust_label)
 
-    @raises_locust_error("uiform.py/click_tab_by_label()")
+    @raises_locust_error
     def click_tab_by_label(self, tab_label: str, tab_group_test_label: str, locust_request_label: str = "") -> 'SailUiForm':
         """
         Selects a Tab by its label and its tab group's testLabel
@@ -690,7 +878,7 @@ class SailUiForm:
 
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/upload_document_to_upload_field()")
+    @raises_locust_error
     def upload_document_to_upload_field(self, label: str, file_path: str, locust_request_label: str = "") -> 'SailUiForm':
         """
         Uploads a document to a named upload field
@@ -738,7 +926,7 @@ class SailUiForm:
                 f"No response returned when trying to upload file to field '{label}'")
         return self._reconcile_state(new_state)
 
-    @raises_locust_error("uiform.py/fill_date_field()")
+    @raises_locust_error
     def fill_date_field(self, label: str, date_input: datetime.date, locust_request_label: str = "") -> 'SailUiForm':
         """
         Fills a date field with the specified date
@@ -776,7 +964,7 @@ class SailUiForm:
 
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/fill_datetime_field()")
+    @raises_locust_error
     def fill_datetime_field(self, label: str, datetime_input: datetime.datetime, locust_request_label: str = "") -> 'SailUiForm':
         """
         Fills a datetime field with the specified datetime
@@ -818,7 +1006,38 @@ class SailUiForm:
 
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/move_to_end_of_paging_grid()")
+    @raises_locust_error
+    def select_rows_in_grid(self, rows: List[int], label: str = None, index: int = None, locust_request_label: str = "") -> 'SailUiForm':
+        """
+        Selects rows in a grid
+        Either a label or an index is required, indices are useful if there is no title for the grid
+
+        Args:
+            rows(List[int]): The rows to select
+            label(str): Label of the grid
+            index(int): Index of the grid
+
+        Keyword Args:
+            locust_request_label(str): Label used to identify the request for locust statistics
+
+        Returns (SailUiForm): The latest state of the UiForm
+
+        Examples:
+
+            >>> form.select_rows_in_grid(rows=[1], label='my nice grid')
+        """
+        grid = self.grid_interactor.find_grid_by_label_or_index(self.state, label=label, index=index)
+        grid_label = self.grid_interactor.format_grid_display_label(grid)
+
+        new_grid_save = self.grid_interactor.select_rows(grid, rows)
+        context_label = locust_request_label or f"{self.breadcrumb}.Grid.SelectRows.{grid_label}"
+
+        reeval_url = self._get_update_url_for_reeval(self.state)
+        new_state = self.interactor.update_grid_from_sail_form(reeval_url, grid, new_grid_save,
+                                                               self.context, self.uuid, context_label=context_label)
+        return self._reconcile_state(new_state, form_url=reeval_url)
+
+    @raises_locust_error
     def move_to_end_of_paging_grid(self, label: str = None, index: int = None, locust_request_label: str = "") -> 'SailUiForm':
         """
         Moves to the end of a paging grid, if possible
@@ -826,7 +1045,7 @@ class SailUiForm:
 
         Args:
             label(str): Label of the grid
-            index(str): Index of the grid
+            index(int): Index of the grid
 
         Keyword Args:
             locust_request_label(str): Label used to identify the request for locust statistics
@@ -848,7 +1067,7 @@ class SailUiForm:
                                                                self.context, self.uuid, context_label=context_label)
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/move_to_beginning_of_paging_grid()")
+    @raises_locust_error
     def move_to_beginning_of_paging_grid(self, label: str = None, index: int = None, locust_request_label: str = "") -> 'SailUiForm':
         """
         Moves to the beginning of a paging grid, if possible
@@ -856,7 +1075,7 @@ class SailUiForm:
 
         Args:
             label(str): Label of the grid
-            index(str): Index of the grid
+            index(int): Index of the grid
 
         Returns (SailUiForm): The latest state of the UiForm
 
@@ -875,7 +1094,7 @@ class SailUiForm:
                                                                self.context, self.uuid, context_label=context_label)
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/move_to_left_in_paging_grid()")
+    @raises_locust_error
     def move_to_left_in_paging_grid(self, label: str = None, index: int = None, locust_request_label: str = "") -> 'SailUiForm':
         """
         Moves to the left in a paging grid, if possible
@@ -884,7 +1103,7 @@ class SailUiForm:
 
         Args:
             label(str): Label of the grid
-            index(str): Index of the grid
+            index(int): Index of the grid
 
         Keyword Args:
             locust_request_label(str): Label used to identify the request for locust statistics
@@ -906,7 +1125,7 @@ class SailUiForm:
                                                                self.context, self.uuid, context_label=context_label)
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/move_to_right_in_paging_grid()")
+    @raises_locust_error
     def move_to_right_in_paging_grid(self, label: str = None, index: int = None, locust_request_label: str = "") -> 'SailUiForm':
         """
         Moves to the right in a paging grid, if possible
@@ -915,7 +1134,7 @@ class SailUiForm:
 
         Args:
             label(str): Label of the grid
-            index(str): Index of the grid
+            index(int): Index of the grid
 
         Keyword Args:
             locust_request_label(str): Label used to identify the request for locust statistics
@@ -937,7 +1156,7 @@ class SailUiForm:
                                                                self.context, self.uuid, context_label=context_label)
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/sort_paging_grid()")
+    @raises_locust_error
     def sort_paging_grid(self, label: str = None, index: int = None, field_name: str = "", ascending: bool = False, locust_request_label: str = "") -> 'SailUiForm':
         """
         Sorts a paging grid by the field name, which is not necessarily the same as the label of the column
@@ -949,7 +1168,7 @@ class SailUiForm:
 
         Args:
             label(str): Label of the grid
-            index(str): Index of the grid
+            index(int): Index of the grid
             field_name(str): Field to sort on (not necessarily the same as the displayed one)
             ascending(bool): Whether to sort ascending, default is false
 
@@ -975,7 +1194,7 @@ class SailUiForm:
                                                                self.context, self.uuid, context_label=context_label)
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/select_radio_button_by_test_label()")
+    @raises_locust_error
     def select_radio_button_by_test_label(self, test_label: str, index: int, locust_request_label: str = "") -> 'SailUiForm':
         """
         Selects a radio button by its test label
@@ -1006,7 +1225,7 @@ class SailUiForm:
                 f"No response returned when trying to select radio button with testLabel '{test_label}'")
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/select_radio_button_by_label()")
+    @raises_locust_error
     def select_radio_button_by_label(self, label: str, index: int, locust_request_label: str = "") -> 'SailUiForm':
         """
         Selects a radio button by its label
@@ -1040,7 +1259,7 @@ class SailUiForm:
                 f"No response returned when trying to select radio button with label '{label}'")
         return self._reconcile_state(new_state, form_url=reeval_url)
 
-    @raises_locust_error("uiform.py/select_radio_button_by_index()")
+    @raises_locust_error
     def select_radio_button_by_index(self, field_index: int, index: int, locust_request_label: str = "") -> 'SailUiForm':
         """
         Selects a radio button by its field index
@@ -1166,10 +1385,84 @@ class SailUiForm:
             raise Exception(f"At least one validation was found in the form {self.breadcrumb}")
         return self
 
+    @raises_locust_error
+    def refresh_after_record_action(self, label: str, locust_request_label: str = "") -> 'SailUiForm':
+        """
+        Refreshes a form after the completion of a record action.
+
+        Args:
+            label(str): Label of the record action that has just been completed
+
+        Keyword Args:
+            locust_request_label(str): Label used to identify the request for locust statistics
+
+        Returns (SailUiForm): The latest state of the UiForm
+
+        Examples:
+
+            >>> initial_form = copy(form)
+            >>> form.click('Request upgrade')
+            >>> ...
+            >>> form.click('Submit')
+            >>> initial_form.refresh_after_record_action('Request upgrade')
+
+        """
+
+        record_action_component = find_component_by_attribute_in_dict(
+            'label', label, self.state)
+        self._validate_component_found(record_action_component, label)
+
+        record_action_trigger_component = find_component_by_attribute_in_dict(
+            '_actionName', 'sail:record-action-trigger', self.state)
+        self._validate_component_found(record_action_trigger_component, label)
+
+        reeval_url = self._get_update_url_for_reeval(self.state)
+        locust_label = locust_request_label or f"{self.breadcrumb}.RefreshAfterRecordAction.{label}"
+        new_state = self.interactor.refresh_after_record_action(
+            reeval_url, record_action_component, record_action_trigger_component, self.context, self.uuid, label=locust_label)
+
+        if not new_state:
+            raise Exception(f"No response returned when trying to refresh after record action '{label}'")
+
+        return self._reconcile_state(new_state, form_url=reeval_url)
+
+    @raises_locust_error
+    def click_record_search_button_by_index(self, index: int = 1, locust_request_label: str = "") -> 'SailUiForm':
+        """
+        Clicks the Search button of a record grid.
+
+        Args:
+            index(int): Index of the record search button on the form
+
+        Keyword Args:
+            locust_request_label(str): Label used to identify the request for locust statistics
+
+        Returns (SailUiForm): The latest state of the UiForm
+
+        Examples:
+
+            >>> form.click_record_search_button_by_index(1)
+
+        """
+        component = find_component_by_index_in_dict("SearchBoxWidget", index, self.state)
+        self._validate_component_found_by_attribute(component, "searchButtonLabel", "Search")
+
+        reeval_url = self._get_update_url_for_reeval(self.state)
+        locust_label = locust_request_label or f"{self.breadcrumb}.ClickRecordSearchButtonByIndex.{index}"
+        new_state = self.interactor.click_record_search_button(
+            reeval_url, component, self.context, self.uuid, label=locust_label)
+
+        if not new_state:
+            raise Exception(f"No response returned when trying to click record search button at index '{index}'")
+
+        return self._reconcile_state(new_state, form_url=reeval_url)
+
     def _reconcile_state(self, new_state: dict, form_url: str = "") -> 'SailUiForm':
         self.interactor.datatype_cache.cache(new_state)
         self.state = self.reconciler.reconcile_ui(self.state, new_state)
         self.form_url = form_url or self.form_url
+        self.uuid = self.state.get(KEY_UUID) or self.uuid
+        self.context = self.state.get(KEY_CONTEXT) or self.context
         return self
 
     def _validate_component_found(self, component: Optional[Dict[str, Any]], label: str, type: Optional[str] = None) -> None:
@@ -1200,15 +1493,3 @@ class SailUiForm:
                 reeval_url = urlparse(link_object.get("href", "")).path
                 break
         return reeval_url
-
-
-class ComponentNotFoundException(Exception):
-    pass
-
-
-class InvalidComponentException(Exception):
-    pass
-
-
-class ChoiceNotFoundException(Exception):
-    pass
